import Boggle_mod
import random

def play_boggle():
    """
    Desc:   Launches the boggle game.
    Input:  None
    Output: None
    """
    dico = Boggle_mod.load_dico("dico.txt")
    points = 0
    for i in range(5):
        points += play_one_round(dico)  
        print()
    print("Vous gagnez", points, "point(s) au total")

def play_one_round(dico):
    """
    Desc:   Play one round of Boggle.
    Input:  dico (set)
    Output: points (int)
    """
    board = shake()
    Boggle_mod.show_grid(board)
    word = ask_word()
    (points, msg) = check_word(word, dico, board)
    show_status(msg)
    return points
    
def show_status(msg):
    """
    Desc:   Shows the status of the game.
    Input:  msg (str)
    Output: None
    """
    print(msg)
  
def shake():
    """
    Desc:   Returns a 4x4 random board formed from shaking the dice.
    Input:  None
    Output: (list(list(str)))
    """
    ds = Boggle_mod.get_dices()
    shuffle(ds)                  # dice in random order
    ls = random_faces(ds)        # list of random faces
    return list_to_board(ls)
  
def shuffle(ls):
    """
    Desc:   Shuffles a list.
    Input:  ls (list)
    Output: None
    """
    random.shuffle(ls)

def random_faces(ds):
    """
    Desc:   Returns a list of random faces.
    Input:  ds (list(str))
    Output: ls (list(str))
    """
    ls = []
    for d in ds:
        ls.append(random_face(d))
    return ls
  
def random_face(dice):
    """
    Desc:   Returns a random face.
    Input:  (str)
    Output: (str)
    """
    return random.choice(dice)
  
def list_to_board(ls):
    """
    Desc:   Returns the board made from shuffled dices
    Input:  ls (list(str))
    Output: board (list(list(str)))
    """
    board = []
    for i in range(0, len(ls), 4):
        board.append(ls[i:i+4])
    return board

def show_board(board):
    """
    Desc:   Prints the 4x4 board.
    Input:  board (list(list(str)))
    Output: None
    """
    for line in board:
        for val in line:
            print(val, end = ' ')
        print()

def ask_word():
    """
    Desc:   Asks and returns the chosen word by the user.
    Input:  None
    Output: word (str)
    """  
    word = ask_input('Quel est votre mot: ')
    while len(word) < 3 or not is_pure_upalpha(word):
        word = ask_input('Proposition invalide. Recommencez\nQuel est votre mot: ')
    return word

def ask_input(msg):
    return input(msg).upper().strip()
    
def is_pure_upalpha(word):
    """
    Desc:   Checks if all characters in word are uppercase alpha.
    Input:  word (str)
    Output: res (bool)
    """
    res = len(word) > 0
    k = 0
    while k < len(word) and res:
        res = 'A' <= word[k] <= 'Z'
        k += 1
    return res
  
def check_word(word, dico, board):
    """
    Desc:   Checks if the word is playable and returns the earned points.
    Input:  word (str), dico (set), board (list(list(str)))
    Output: result (tuple(int, str))
    """
    if not word_in_dico(word, dico):
        result = (0, "Mot absent du dictionnaire. Vous ne gagnez pas de points")
    elif not word_on_board(word, board):
        result = (0, "Mot introuvable. Vous ne gagnez pas de points")
    else:   
        points = check_points(word)
        result = (points, "Vous gagnez " + str(points) + " points")
    return result

def word_in_dico(word, dico):
    """
    Desc:   Checks if a word exists in a dictionary.
    Input:  word (str), dico (set)
    Output: (bool)
    """
    return word in dico

def word_on_board(word, board):
    """
    Desc:   Checks if a word is on the board, according to the Boggle's rules.
    Input:  word (str), board (list(list(str)))
    Output: found (bool)
    """
    found = False
    line = 0
    while line < 4 and not found:
        found = word_starts_on_line(word, board, line)
        line += 1
    return found
  
def word_starts_on_line(word, board, line):
    """
    Desc:   Checks if a word is on a line.
    Input:  word (str), board (list(list(str))), line (int)
    Output: found (bool)
    """
    found = False
    col = 0
    while col < 4 and not found:
        found = word_in_place(word, board, line, col)
        col += 1
    return found

def word_in_place(word, board, line, col):
    """
    Desc:   Checks if the word on the board is starting at the position (line, col).
    Input:  word (str), board (list(list(str))), line (int), col (int)
    Output: found (bool)
    """
    cboard = deep_copy(board)
    found = letter_in_place(word[0], cboard, line, col)
    k = 1
    while k < len(word) and found:
        (found, line, col) = letter_in_neighbourhood(word[k], cboard, line, col)
        k += 1
    return found

def deep_copy(board):
    """
    Desc:   Makes a deep copy of a list of list.
    Input:  board (list(list(str)))
    Output: copy (list(list(str)))
    """
    copy = []
    for l in board:
        copy.append(l[:])
    return copy

def letter_in_place(letter, board, line, col):
    """
    Desc:   Checks if a letter appear on the board at the position (line, col).
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: found (bool)
    """
    found = board[line][col] == letter
    if found:
        board[line][col] = '*' # this place marked as used
    return found
  
def letter_in_neighbourhood(letter, board, line, col):
    """
    Desc:   Marks the letter if it is a neighbour of (line, col).
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: (tup(bool, int, int))
    """
    (found, line, col) = check_neighbours(letter, board, line, col)
    if found:
        board[line][col] = '*' # this place marked as used
    return (found, line, col)

def check_neighbours(letter, board, line, col):
    """
    Desc:   Checks if a letter appear in the neighbourhood of the position (line, col) and
            returns the possible  of a possible occurence of letter as a neighbour of (line, col) on the board.
            Returns (False, -1, -1) if letter in not a neighbour of (line, col).
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: (tup(bool, int, int))
    """
    if is_north(letter, board, line, col):
        line -= 1
    elif is_south(letter, board, line, col):
        line += 1
    elif is_west(letter, board, line, col):
        col -= 1
    elif is_east(letter, board, line, col):
        col += 1
    else:
        (line, col) = (-1, -1)
    return (is_valid_neighbour((line, col)), line, col)
  
def is_north(letter, board, line, col):
    """
    Desc:   Checks if the letter appears north of the position (line, col) on the board.
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: (bool)
    """
    return line > 0 and board[line - 1][col] == letter           

def is_south(letter, board, line, col):
    """
    Desc:   Checks if the letter appears south of the position (line, col) on the board.
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: (bool)
    """
    return line < 3 and board[line + 1][col] == letter
            
def is_west(letter, board, line, col):
    """
    Desc:   Checks if the letter appears west of the position (line, col) on the board.
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: (bool)
    """
    return col > 0 and board[line][col - 1] == letter
           
def is_east(letter, board, line, col):
    """
    Desc:   Checks if the letter appears east of the position (line, col) on the board.
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: (bool)
    """
    return col < 3 and board[line][col + 1] == letter
           
def is_valid_neighbour(pos):
    """
    Desc:   Checks if the position indicates a valid neighbour (not (-1, -1)).
    Input:  pos (tuple(int, int))
    Output: (bool)
    """
    return pos != (-1, -1)
  
def check_points(word):
    """
    Desc:   Returns the points according the the word's length.
    Input:  word (str)
    Output: point (int)
    """
    lg = len(word)
    if lg < 8:
        points = [0,0,0,1,1,2,3,5][lg]
    else:
        points = 11
    return points

if __name__ == "__main__":
    play_boggle()

