import Boggle_mod
import random

def play_boggle():
    """
    Desc:   Launches the Boggle game.
    Input:  none
    Output: none
    """
    dico = Boggle_mod.load_dico("dico.txt")
    play = True
    while play:
        points = 0
        for i in range(5):
            points += play_one_round(dico)  
            print()
            print("Vous gagnez", points, "point(s) au total")
        play = Boggle_mod.ask_yes_no("\nVoulez-vous jouer une nouvelle partie?")
        print()

def play_one_round(dico):
    """
    Desc:   Play one round of Boggle.
    Input:  (set) dico -- dictionary of valid words
    Output: (int) -- the points gained by the player during the round
    """
    board = shake()
    Boggle_mod.show_grid(board)
    word = ask_word()
    (points, msg) = play_word(word, dico, board)
    print(msg)
    return points
      
def shake():
    """
    Desc:   Returns a 4x4 random board formed from shaking the dices.
    Input:  none
    Output: (list(list(str))) -- the shuffled board
    """
    ds = Boggle_mod.get_dices()
    random.shuffle(ds)
    ls = []
    for d in ds:
        ls.append(random.choice(d))
    return list_to_board(ls)
      
def list_to_board(ls):
    """
    Desc:   Transforms a list of 16 values into a 4x4 board
    Input:  (list(str)) ls -- list of values
    Output: (list(list(str))) -- the board
    """
    board = []
    for i in range(0, len(ls), 4):
        board.append(ls[i:i+4])
    return board

def ask_word():
    """
    Desc:   Asks and returns a word from the user.
    Input:  none
    Output: (str) word -- a Boggle rule-compliant word
    """  
    word = ask_input('Quel est votre mot: ')
    while len(word) < 3 or not is_pure_upalpha(word):
        word = ask_input('Proposition invalide. Recommencez\nQuel est votre mot: ')
    return word

def ask_input(msg):
    """
    Desc:   Reads a user input from the keyboard.
    Input:  (str) msg -- the prompt message to be displayed
    Output: (str) uppercased and stripped input provided by the user
    """
    return input(msg).upper().strip()
    
def is_pure_upalpha(word):
    """
    Desc:   Checks if all characters in word are uppercase alpha.
    Input:  (str) word -- word to be checked
    Output: (bool)
    """
    res = len(word) > 0
    k = 0
    while k < len(word) and res:
        res = 'A' <= word[k] <= 'Z'
        k += 1
    return res
  
def play_word(word, dico, board):
    """
    Desc:   Plays a word against a dictionary and a board.
    Input:  (str) word -- word to be checked, 
            (set) dico -- dictionary of valid words,
            (list(list(str))) board -- board of letters to search the word on
    Output: (tuple(int, str)) (points, msg) -- points earned and message to be displayed
    """
    if not word in dico:
        result = (0, "Mot absent du dictionnaire. Vous ne gagnez pas de points")
    elif not word_on_board(word, board):
        result = (0, "Mot introuvable. Vous ne gagnez pas de points")
    else:   
        points = check_points(word)
        result = (points, "Vous gagnez " + str(points) + " points")
    return result

def word_on_board(word, board):
    """
    Desc:   Checks if a word is on the board, according to Boggle's rules.
    Input:  (str) word -- word to be checked, 
            (list(list(str))) board -- board of letters to search the word on
    Output: (bool)
    """
    found = False
    line = 0
    while line < 4 and not found:
        col = 0
        while col < 4 and not found:
            found = word_in_place(word, board, line, col)
            col += 1
        line += 1
    return found
  
def word_in_place(word, board, line, col):
    """
    Desc:   Checks if the word in on the board, starting at position (line, col).
    Input:  (str) word -- word to be checked, 
            (list(list(str))) board -- board of letters to search the word on, 
            (int) line, (int) col -- position on the board where to start the search
    Output: (bool)
    """
    visited = []
    is_ok = (board[line][col] == word[0])
    k = 1
    while k < len(word) and is_ok:
        visited.append((line,col))
        (found, line, col) = find_next_letter(word[k], board, line, col)
        is_ok = found and (line,col) not in visited
        k += 1
    return is_ok
  
def find_next_letter(letter, board, line, col):
    """
    Desc:   Find a letter on the board in a position adjacent to (line,col)
    Input:  letter (str), board (list(list(str))), line (int), col (int)
    Output: (tup(bool, int, int)) (found, line, col)
                   found -- True if letter has been found in an adjacent position; 
                   (line,col) -- position of the letter on the board (if found is True)
    """
    found = True
    if line > 0 and board[line - 1][col] == letter: # check north
        line -= 1
    elif line < 3 and board[line + 1][col] == letter: # check south
        line += 1
    elif col > 0 and board[line][col - 1] == letter: # check west
        col -= 1
    elif col < 3 and board[line][col + 1] == letter: # check east
        col += 1
    else:
        found = False
    return (found, line, col)
             
def check_points(word):
    """
    Desc:   Returns the points according the the word's length.
    Input:  (str) word -- word to be checked
    Output: (int) -- points associated to word
    """
    lg = len(word)
    if lg < 8:
        points = [0,0,0,1,1,2,3,5][lg]
    else:
        points = 11
    return points

if __name__ == "__main__":
    play_boggle()

